# IfcPatch - IFC patching utiliy
# Copyright (C) 2020-2022 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.

import ifcopenshell
import ifcopenshell.api
import ifcopenshell.util.element
from logging import Logger


class Patcher:
    def __init__(self, src: str, file: ifcopenshell.file, logger: Logger, attribute: str = "Tag"):
        """Merge duplicate element types via the Tag or another attribute

        Revit is notorious for creating many duplicate element types. Element
        types may be duplicated by being mirrored, such as doors, columns, etc,
        or being certain MEP equipment. This means that even though you think
        you might have only 3 door families and 3 door types in your door
        schedule, your IFC might actually incorrectly store 6 or more door types.

        Revit stores the Revit Element ID in the "Tag" attribute of all IFC
        elements, so we can deduce that multiple IFC elements with the same Tag
        attribute have been duplicated in IFC. This patch will merge them into a
        single element type.

        You may optionally specify your own attribute if you want to merge using
        different criteria, such as "Name". For example, may Revit users
        incorrectly have multiple types with the same name as workarounds to
        overcome various Revit limitations. This is incorrect and this patch
        will merge the types into a single type.

        Occurrences of the type will be remapped to the merged type.

        :param attribute: The name of the attribute to merge element types based
            on. Typically this will be "Tag" as it stores the unique ID from the
            proprietary BIM software.
        :type attribute: str

        Example:

        .. code:: python

            # Default behaviour of merging by Tag attribute
            ifcpatch.execute({"file": model, "recipe": "MergeDuplicateTypes", "arguments": []})

            # Explicitly say we want to merge based on the Name attribute
            ifcpatch.execute({"file": model, "recipe": "MergeDuplicateTypes", "arguments": ["Name"]})
        """
        self.src = src
        self.file = file
        self.logger = logger
        self.attribute = attribute

    def patch(self):
        key = self.attribute
        keys = {}
        for element_type in self.file.by_type("IfcTypeObject"):
            original_type = keys.get(getattr(element_type, key), None)
            if original_type:
                elements = ifcopenshell.util.element.get_types(element_type)
                if elements:
                    self.assign_type(elements, original_type)
                for inverse in self.file.get_inverse(element_type):
                    ifcopenshell.util.element.replace_attribute(inverse, element_type, original_type)
                self.file.remove(element_type)
            else:
                keys[getattr(element_type, key)] = element_type

    def assign_type(
        self, related_objects: list[ifcopenshell.entity_instance], relating_type: ifcopenshell.entity_instance
    ) -> None:
        # To be conservative, we disable `should_map_representations`
        # since that would do other things like
        # map type representations or recalculate material set usages which is
        # risky when we're patching an existing dataset.
        ifcopenshell.api.run(
            "type.assign_type",
            self.file,
            relating_type=relating_type,
            related_objects=related_objects,
            should_map_representations=False,
        )



# IfcOpenShell - IFC toolkit and geometry engine
# Copyright (C) 2022 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcOpenShell.
#
# IfcOpenShell is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcOpenShell is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcOpenShell.  If not, see <http://www.gnu.org/licenses/>.

import os
import pytest
import ifcpatch
import ifcopenshell
import ifcopenshell.api
import ifcopenshell.util.element
import test.bootstrap


class TestRegenerateGlobalIds(test.bootstrap.IFC4):
    def test_run(self):
        project = ifcopenshell.api.run("root.create_entity", self.file, ifc_class="IfcProject")
        wall = ifcopenshell.api.run("root.create_entity", self.file, ifc_class="IfcWall")
        used_guids = {project.GlobalId, wall.GlobalId}
        ifcpatch.execute({"file": self.file, "recipe": "RegenerateGlobalIds", "arguments": [False]})
        new_guids = {project.GlobalId, wall.GlobalId}
        assert not new_guids.intersection(used_guids)

    def test_regenerate_guids_for_duplicates(self):
        project = ifcopenshell.api.run("root.create_entity", self.file, ifc_class="IfcProject")
        wall1 = ifcopenshell.api.run("root.create_entity", self.file, ifc_class="IfcWall")
        wall2 = ifcopenshell.api.run("root.create_entity", self.file, ifc_class="IfcWall")
        wall1.GlobalId = wall2.GlobalId
        used_guids = {project, wall1.GlobalId, wall2.GlobalId}

        ifcpatch.execute({"file": self.file, "recipe": "RegenerateGlobalIds", "arguments": [True]})
        new_guids = {project, wall1.GlobalId, wall2.GlobalId}
        assert len(new_guids) == 3
        assert len(new_guids.intersection(used_guids)) == 2


class TestRegenerateGlobalIdsIFC2X3(test.bootstrap.IFC2X3, TestRegenerateGlobalIds):
    pass



# IfcPatch - IFC patching utiliy
# Copyright (C) 2020, 2021 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.

import ifcopenshell
import ifcopenshell.util.element


class Patcher:
    def __init__(self, src, file, logger):
        """Optimise the filesize of an IFC model

        It is possible to non-losslessly optimise the filesize of an IFC model.

        Note that this is usually not recommended. Optimising runs a risk of
        losing some indirect semantic data critical for native IFC authoring.
        Most parties who recommend optimisation are not aware of these risks.
        Optimising is only safe in the context of read-only IFCs.

        If filesize is an issue, another approach would be to use IFCZIP
        instead to compress the model. Optimising the model only typically
        affects filesize and has minimal impact on load times. Large filesizes
        can usually be solved through other means. Consult the BlenderBIM Add-on
        documentation on dealing with large models for more details.

        Warning: this optimise recipe is very, very slow. Please consider using
        RecycleNonRootedElements instead.

        Example:

        .. code:: python

            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "Optimise", "arguments": []})
        """
        self.src = src
        self.file = file
        self.logger = logger
        self.optimized_file = ifcopenshell.file(schema=self.file.schema)

    def patch(self):
        from toposort import toposort_flatten as toposort

        def generate_instances_and_references():
            """
            Generator which yields an entity id and
            the set of all of its references contained in its attributes.
            """
            for inst in self.file:
                yield inst.id(), set(i.id() for i in self.file.traverse(inst)[1:] if i.id())

        instance_mapping = {}

        def map_value(v):
            """
            Recursive function which replicates an entity instance, with
            its attributes, mapping references to already registered
            instances. Indeed, because of the toposort we know that
            forward attribute value instances are mapped before the instances
            that reference them.
            """
            if isinstance(v, (list, tuple)):
                # lists are recursively traversed
                return type(v)(map(map_value, v))
            elif isinstance(v, ifcopenshell.entity_instance):
                if v.id() == 0:
                    # express simple types are not part of the toposort and just copied
                    return self.optimized_file.create_entity(v.is_a(), v[0])

                return instance_mapping[v]
            else:
                # a plain python value can just be returned
                return v

        info_to_id = {}

        for id in toposort(dict(generate_instances_and_references())):
            inst = self.file[id]
            info = inst.get_info(include_identifier=False, recursive=True, return_type=frozenset)
            if info in info_to_id:
                mapped = instance_mapping[inst] = instance_mapping[self.file[info_to_id[info]]]

            else:
                info_to_id[info] = id
                instance_mapping[inst] = self.optimized_file.create_entity(inst.is_a(), *map(map_value, inst))

        self.file = self.optimized_file



# IfcOpenShell - IFC toolkit and geometry engine
# Copyright (C) 2022 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcOpenShell.
#
# IfcOpenShell is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcOpenShell is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcOpenShell.  If not, see <http://www.gnu.org/licenses/>.

import ifcpatch
import ifcopenshell
import ifcopenshell.api
import ifcopenshell.util.placement
import test.bootstrap
import tempfile
import numpy as np
from pathlib import Path
from typing import Optional


class TestMergeProject(test.bootstrap.IFC4):
    def setup_project(self, ifc_file: Optional[ifcopenshell.file] = None):
        prefix = None if ifc_file else "MILLI"
        if ifc_file is None:
            ifc_file = ifcopenshell.file(schema=self.file.schema)

        project = ifcopenshell.api.run("root.create_entity", ifc_file, ifc_class="IfcProject")
        unit = ifcopenshell.api.run("unit.add_si_unit", ifc_file, unit_type="LENGTHUNIT", prefix=prefix)
        ifcopenshell.api.run("unit.assign_unit", ifc_file, units=[unit])

        matrix = np.eye(4)
        matrix[:, 3] = (1, 2, 3, 1)
        wall = ifcopenshell.api.run("root.create_entity", ifc_file, ifc_class="IfcWall")
        ifcopenshell.api.run("geometry.edit_object_placement", ifc_file, product=wall, matrix=matrix, is_si=True)
        return ifc_file

    def test_run(self):
        self.file = self.setup_project(self.file)
        second_file = self.setup_project()
        temp_path = Path(tempfile.gettempdir()) / "second.ifc"
        second_file.write(temp_path)
        output = ifcpatch.execute({"file": self.file, "recipe": "MergeProject", "arguments": [str(temp_path)]})

        assert len(output.by_type("IfcWall")) == 2
        wall1, wall2 = output.by_type("IfcWall")

        # test that units are converted
        placement1 = ifcopenshell.util.placement.get_local_placement(wall1.ObjectPlacement)
        placement2 = ifcopenshell.util.placement.get_local_placement(wall2.ObjectPlacement)
        to_tuple = lambda arr: tuple(map(tuple, arr))
        matrix = np.eye(4)
        matrix[:, 3] = (1, 2, 3, 1)
        assert to_tuple(placement1) == to_tuple(placement2) == to_tuple(matrix)


class TestMergeProjectIFC2X3(test.bootstrap.IFC2X3, TestMergeProject):
    pass



# IfcPatch - IFC patching utiliy
# Copyright (C) 2023 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.

import math
import numpy as np
import ifcopenshell


class Patcher:
    def __init__(self, src, file, logger, x=None, y=None, z=None, ax=None, ay=None, az=None):
        """Sets the world coordinate system of the geometric representation context

        Sets the world coordinate system to whatever you want.

        :param x: The X coordinate.
        :type x: float
        :param y: The Y coordinate.
        :type y: float
        :param z: The Z coordinate.
        :type z: float
        :param ax: An angle to rotate by for 3D rotations along the X axis.
            Angles are in decimal degrees and positive is anticlockwise.
        :type ax: float
        :param ay: An angle to rotate by for 3D rotations along the Y axis.
            Angles are in decimal degrees and positive is anticlockwise.
        :type ay: float
        :param az: An angle to rotate by for 3D rotations along the Z axis.
            Angles are in decimal degrees and positive is anticlockwise.
        :type az: float

        Example:

        .. code:: python

            # Set the world coordinate system back to 0, 0, 0
            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "SetWorldCoordinateSystem", "arguments": [0,0,0,0,0,0]})
        """
        self.src = src
        self.file = file
        self.logger = logger
        self.x = x
        self.y = y
        self.z = z
        self.ax = ax
        self.ay = ay
        self.az = az

    def patch(self):
        rotate = self.identity_matrix()

        for arg in (("x", float(self.ax)), ("y", float(self.ay)), ("z", float(self.az))):
            if arg[1]:
                rotate = getattr(self, f"{arg[0]}_rotation_matrix")(math.radians(arg[1]), rotate)

        m = rotate

        x = np.array((m[0][0], m[1][0], m[2][0]))
        z = np.array((m[0][2], m[1][2], m[2][2]))

        for context in self.file.by_type("IfcGeometricRepresentationContext", include_subtypes=False):
            if context.ContextType == "Model":
                context.WorldCoordinateSystem = self.file.createIfcAxis2Placement3D(
                    self.file.createIfcCartesianPoint((float(self.x), float(self.y), float(self.z))),
                    self.file.createIfcDirection(z.tolist()),
                    self.file.createIfcDirection(x.tolist()),
                )
            elif context.ContextType == "Plan":
                context.WorldCoordinateSystem = self.file.createIfcAxis2Placement2D(
                    self.file.createIfcCartesianPoint((float(self.x), float(self.y), float(self.z))),
                    self.file.createIfcDirection(x.tolist()),
                )

    def identity_matrix(self):
        return np.eye(4)

    def x_rotation_matrix(self, angle, transformation):
        transformation[1][1] = math.cos(angle)
        transformation[1][2] = -math.sin(angle)
        transformation[2][1] = math.sin(angle)
        transformation[2][2] = math.cos(angle)
        return transformation

    def y_rotation_matrix(self, angle, transformation):
        transformation[0][0] = math.cos(angle)
        transformation[0][2] = math.sin(angle)
        transformation[2][0] = -math.sin(angle)
        transformation[2][2] = math.cos(angle)
        return transformation

    def z_rotation_matrix(self, angle, transformation):
        transformation[0][0] = math.cos(angle)
        transformation[0][1] = -math.sin(angle)
        transformation[1][0] = math.sin(angle)
        transformation[1][1] = math.cos(angle)
        return transformation



# IfcPatch - IFC patching utiliy
# Copyright (C) 2023 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.

import ifcopenshell
import ifcopenshell.util.element


class Patcher:
    def __init__(self, src, file, logger):
        """Reassigns occurrence classifications to types

        Revit has a bug (see https://github.com/Autodesk/revit-ifc/issues/691)
        where it assigns classification codes to occurrences instead of types.
        Almost always, this is not what you want. This patch reassigns all
        classifications to their respective types.

        Example:

        .. code:: python

            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "FixRevitClassificationCodeTypes"})
        """
        self.src = src
        self.file = file
        self.logger = logger

    def patch(self):
        for rel in self.file.by_type("IfcRelAssociatesClassification"):
            related_types = set()
            for related_object in rel.RelatedObjects:
                relating_type = ifcopenshell.util.element.get_type(related_object)
                if relating_type:
                    related_types.add(relating_type)
                else:
                    related_types.add(related_object)
            rel.RelatedObjects = list(related_types)



# IfcPatch - IFC patching utiliy
# Copyright (C) 2020, 2021 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.

import ifcopenshell
import ifcopenshell.util.pset
import ifcopenshell.util.element


class Patcher:
    def __init__(self, src, file, logger, property_name=None, quantity_name=None):
        """Converts a property to a standardised quantity

        IFC can store arbitrary key value metadata associated with a elements
        known as properties and quantities. The difference between the two is
        that quantities specifically measure physical dimensions, and can be
        used parametrically. A collection of standardised quantities have also
        been published by buildingSMART. Using standardised quantities can help
        automate BIM workflows, instead of arbitrary properties.

        Sometimes, proprietary BIM software incorrectly stores quantities and
        properties. This patch lets you convert these incorrect properties to
        standardised quantities instead.

        The existing property will not be removed.

        :param property_name: The name of the property to convert into a
            quantity. The name of the property set is not considered.
        :type property_name: str
        :param quantity_name: The name of the quantity that this property should
            be stored in. This should be a standard name that is one of the
            quantity names of a buildingSMART quantity template. For example, it
            may be "NetSideArea" for walls, which exists in
            Qto_WallBaseQuantities. The quantity set name will be based on the
            standard buildingSMART quantity template.
        :type quantity_name: str

        Example:

        .. code:: python

            # Any property named "Area" is converted to a quantity named
            # "NetSideArea", if that standardised quantity exists.
            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "ConvertPropertiesToQuantities", "arguments": ["Area", "NetSideArea"]})
        """
        self.src = src
        self.file = file
        self.logger = logger
        self.source_property_name = property_name
        self.destination_quantity_name = quantity_name

    def patch(self):
        self.qto_template_cache = {}
        self.psetqto = ifcopenshell.util.pset.get_template("IFC4")

        for product in self.file.by_type("IfcTypeProduct"):
            self.process_product(product, product.HasPropertySets or [])

        for product in self.file.by_type("IfcProduct"):
            self.process_product(
                product,
                [r.RelatingPropertyDefinition for r in product.IsDefinedBy if r.is_a("IfcRelDefinesByProperties")],
            )

    def process_product(self, product, definitions):
        value = None
        has_quantity = False
        qtos = {}
        for definition in definitions or []:
            if definition.is_a("IfcPropertySet"):
                for prop in definition.HasProperties:
                    if prop.is_a("IfcPropertySingleValue") and prop.Name == self.source_property_name:
                        value = prop.NominalValue.wrappedValue if prop.NominalValue else None
            elif definition.is_a("IfcElementQuantity"):
                qtos[definition.Name] = definition
                for quantity in definition.Quantities:
                    if quantity.is_a("IfcPhysicalSimpleQuantity") and quantity.Name == self.destination_quantity_name:
                        has_quantity = True

        if value and not has_quantity:
            qto_name = self.get_qto_name(product.is_a())
            qto = qtos.get(qto_name, ifcopenshell.api.run("pset.add_qto", self.file, product=product, Name=qto_name))
            ifcopenshell.api.run(
                "pset.edit_qto", self.file, qto=qto, Properties={self.destination_quantity_name: value}
            )

    def get_qto_name(self, ifc_class):
        for template in self.get_qto_templates(ifc_class):
            if self.destination_quantity_name in [t.Name for t in template.HasPropertyTemplates]:
                return template.Name

    def get_qto_templates(self, ifc_class):
        if ifc_class not in self.qto_template_cache:
            self.qto_template_cache[ifc_class] = self.psetqto.get_applicable(ifc_class, qto_only=True)
        return self.qto_template_cache[ifc_class]



# IfcPatch - IFC patching utiliy
# Copyright (C) 2020, 2021 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.

import ifcopenshell
import ifcopenshell.util.schema
from logging import Logger


class Patcher:
    def __init__(self, src: str, file: ifcopenshell.file, logger: Logger, schema: str = "IFC4"):
        """Migrate from one IFC version to another

        Note that this is experimental and will try to preserve as much data as
        possible. Upgrading to IFC4 is more stable than downgrading to IFC2X3.

        :param schema: The schema identifier of the IFC version to migrate to.
        :type schema: str

        Example:

        .. code:: python

            # Upgrade an IFC2X3 model to IFC4
            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "Migrate", "arguments": ["IFC4"]})
        """
        self.src = src
        self.file = file
        self.logger = logger
        self.schema = schema

    def patch(self):
        self.file_patched = ifcopenshell.file(schema=self.schema)
        migrator = ifcopenshell.util.schema.Migrator()
        for element in self.file:
            migrator.migrate(element, self.file_patched)
            print("Migrating", element)
            print("Successfully converted to", migrator.migrate(element, self.file_patched))



# IfcPatch - IFC patching utiliy
# Copyright (C) 2020, 2021 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.


class Patcher:
    def __init__(self, src, file, logger, output_dir=None):
        """Split an IFC model into multiple models based on building storey

        The new IFC model names will be named after the storey name in the
        format of {i}-{name}.ifc, where {i} is an ascending number starting from
        0 and {name} is the name of the storey.

        :param output_dir: Specifies an output directory where the new IFC models will be saved.
        :type output_dir: str

        Example:

        .. code:: python

            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "SplitByBuildingStorey", "arguments": ["C:/.../output_files"]})
        """
        self.src = src
        self.file = file
        self.logger = logger
        self.output_dir = output_dir

    def patch(self):
        import ifcopenshell
        from shutil import copyfile

        storeys = self.file.by_type("IfcBuildingStorey")
        for i, storey in enumerate(storeys):
            dest = (
                "{}-{}.ifc".format(i, storey.Name)
                if self.output_dir == None
                else "{}/{}-{}.ifc".format(self.output_dir, i, storey.Name)
            )
            copyfile(self.src, dest)
            old_ifc = ifcopenshell.open(dest)
            new_ifc = ifcopenshell.file(schema=self.file.schema)
            if self.file.schema == "IFC2X3":
                elements = old_ifc.by_type("IfcProject") + old_ifc.by_type("IfcProduct")
            else:
                elements = old_ifc.by_type("IfcContext") + old_ifc.by_type("IfcProduct")
            inverse_elements = []
            for element in elements:
                if element.is_a("IfcElement") and not self.is_in_storey(element, storey):
                    element.Representation = None
                    continue
                if element.is_a("IfcElement"):
                    styled_rep_items = [
                        i for i in old_ifc.traverse(element) if i.is_a("IfcRepresentationItem") and i.StyledByItem
                    ]
                    [new_ifc.add(i.StyledByItem[0]) for i in styled_rep_items]
                new_ifc.add(element)
                inverse_elements.extend(old_ifc.get_inverse(element))
            for inverse_element in inverse_elements:
                new_ifc.add(inverse_element)
            for element in new_ifc.by_type("IfcElement"):
                if not self.is_in_storey(element, storey):
                    new_ifc.remove(element)
            new_ifc.write(dest)

    def is_in_storey(self, element, storey):
        return (
            element.ContainedInStructure
            and element.ContainedInStructure[0].RelatingStructure.is_a("IfcBuildingStorey")
            and element.ContainedInStructure[0].RelatingStructure.GlobalId == storey.GlobalId
        )



# IfcPatch - IFC patching utiliy
# Copyright (C) 2023 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.


import os
import re
import json
import time
import tempfile
import itertools
import numpy as np
import multiprocessing
import ifcopenshell
import ifcopenshell.geom
import ifcopenshell.util.unit
import ifcopenshell.util.shape
import ifcopenshell.util.schema
import ifcopenshell.util.attribute
import ifcopenshell.util.placement

try:
    import sqlite3
except:
    print("No SQLite support")

try:
    import mysql.connector
except:
    print("No MySQL support")


class Patcher:
    def __init__(
        self,
        src,
        file,
        logger,
        sql_type: str = "sqlite",
        host: str = "localhost",
        username: str = "root",
        password: str = "pass",
        database: str = "test",
    ):
        """Convert an IFC-SPF model to SQLite or MySQL.

        There are certain controls which are hardcoded in this recipe that you
        may modify, including:

        - full_schema: if True, will create tables for all IFC classes,
          regardless if they are used or not in the dataset. If False, will
          only create tables for classes in the dataset.
        - is_strict: whether or not to enforce null or not null. If your
          dataset might contain invalid data, set this to False.
        - should_expand: if True, entities with attributes containing lists of
          entities will be separated into multiple rows. This means the ifc_id
          is no longer a unique primary key. If False, lists will be stored as
          JSON.
        - should_get_psets: if True, a separate psets table will be created to
          make it easy to query properties. This is in addition to regular IFC
          tables like IfcPropertySet.
        - should_get_geometry: Whether or not to process and store explicit
          geometry data as a blob in a separate geometry and shape table.
        - should_skip_geometry_data: Whether or not to also create tables for
          IfcRepresentation and IfcRepresentationItem classes. These tables are
          unnecessary if you are not interested in geometry.

        :param sql_type: Choose between "sqlite" or "mysql"
        :type sql_type: str

        Example:

        .. code:: python

            # Convert to SQLite
            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "Ifc2Sql", "arguments": ["sqlite"]})
        """
        self.src = src
        self.file = file
        self.logger = logger
        self.sql_type = sql_type
        self.host = host
        self.username = username
        self.password = password
        self.database = database

    def patch(self):
        self.full_schema = True  # Set true for ifcopenshell.sqlite
        self.is_strict = False
        self.should_expand = False  # Set false for ifcopenshell.sqlite
        self.should_get_inverses = True  # Set true for ifcopenshell.sqlite
        self.should_get_psets = True
        self.should_get_geometry = True  # Set true for ifcopenshell.sqlite
        self.should_skip_geometry_data = False  # Set false for ifcopenshell.sqlite

        self.schema = ifcopenshell.ifcopenshell_wrapper.schema_by_name(self.file.schema)

        if self.sql_type == "sqlite":
            tmp = tempfile.NamedTemporaryFile(delete=False)
            db_file = tmp.name
            self.db = sqlite3.connect(db_file)
            self.c = self.db.cursor()
            self.file_patched = db_file
        elif self.sql_type == "mysql":
            self.db = mysql.connector.connect(
                host=self.host, user=self.username, password=self.password, database=self.database
            )
            self.c = self.db.cursor()
            self.file_patched = None

        self.create_id_map()
        self.create_metadata()

        if self.should_get_psets:
            self.create_pset_table()

        if self.should_get_geometry:
            self.create_geometry_table()
            self.create_geometry()

        if self.full_schema:
            ifc_classes = [d.name() for d in self.schema.declarations() if str(d).startswith("<entity")]
        else:
            ifc_classes = self.file.wrapped_data.types()

        for ifc_class in ifc_classes:
            declaration = self.schema.declaration_by_name(ifc_class)

            if self.should_skip_geometry_data:
                if ifcopenshell.util.schema.is_a(declaration, "IfcRepresentation") or ifcopenshell.util.schema.is_a(
                    declaration, "IfcRepresentationItem"
                ):
                    continue

            if self.sql_type == "sqlite":
                self.create_sqlite_table(ifc_class, declaration)
            elif self.sql_type == "mysql":
                self.create_mysql_table(ifc_class, declaration)
            self.insert_data(ifc_class)

        if self.should_get_geometry:
            if self.sql_type == "sqlite":
                if self.shape_rows:
                    self.c.executemany("INSERT INTO shape VALUES (?, ?, ?, ?, ?, ?);", self.shape_rows.values())
                if self.geometry_rows:
                    self.c.executemany("INSERT INTO geometry VALUES (?, ?, ?, ?, ?, ?);", self.geometry_rows.values())
            elif self.sql_type == "mysql":
                if self.shape_rows:
                    self.c.executemany("INSERT INTO shape VALUES (%s, %s, %s, %s, %s, %s);", self.shape_rows.values())
                # Do row by row in case of max_allowed_packet
                for row in self.geometry_rows.values():
                    self.c.execute("INSERT INTO geometry VALUES (%s, %s, %s, %s, %s, %s);", row)

        self.db.commit()
        self.db.close()

    def create_geometry(self):
        self.unit_scale = ifcopenshell.util.unit.calculate_unit_scale(self.file)

        self.shape_rows = {}
        self.geometry_rows = {}

        if self.file.schema in ("IFC2X3", "IFC4"):
            self.elements = self.file.by_type("IfcElement") + self.file.by_type("IfcProxy")
        else:
            self.elements = self.file.by_type("IfcElement")

        self.settings = ifcopenshell.geom.settings()
        self.settings.set(self.settings.STRICT_TOLERANCE, True)

        self.body_contexts = [
            c.id()
            for c in self.file.by_type("IfcGeometricRepresentationSubContext")
            if c.ContextIdentifier in ["Body", "Facetation"]
        ]
        # Ideally, all representations should be in a subcontext, but some BIM programs don't do this correctly
        self.body_contexts.extend(
            [
                c.id()
                for c in self.file.by_type("IfcGeometricRepresentationContext", include_subtypes=False)
                if c.ContextType == "Model"
            ]
        )
        self.settings.set_context_ids(self.body_contexts)

        products = self.elements
        iterator = ifcopenshell.geom.iterator(self.settings, self.file, multiprocessing.cpu_count(), include=products)
        valid_file = iterator.initialize()
        checkpoint = time.time()
        progress = 0
        total = len(products)
        while True:
            progress += 1
            if progress % 250 == 0:
                percent_created = round(progress / total * 100)
                percent_preprocessed = iterator.progress()
                percent_average = (percent_created + percent_preprocessed) / 2
                print(
                    "{} / {} ({}% created, {}% preprocessed) elements processed in {:.2f}s ...".format(
                        progress, total, percent_created, percent_preprocessed, time.time() - checkpoint
                    )
                )
                checkpoint = time.time()
            shape = iterator.get()
            if shape:
                if shape.geometry.id not in self.geometry_rows:
                    v = np.array(shape.geometry.verts).tobytes()
                    e = np.array(shape.geometry.edges).tobytes()
                    f = np.array(shape.geometry.faces).tobytes()
                    mids = np.array(shape.geometry.material_ids).tobytes()
                    m = json.dumps([int(m.name.split("-")[2]) for m in shape.geometry.materials])
                    self.geometry_rows[shape.geometry.id] = [shape.geometry.id, v, e, f, mids, m]
                m = ifcopenshell.util.shape.get_shape_matrix(shape)
                m[0][3] /= self.unit_scale
                m[1][3] /= self.unit_scale
                m[2][3] /= self.unit_scale
                x, y, z = m[:, 3][0:3]
                self.shape_rows[shape.id] = [shape.id, float(x), float(y), float(z), m.tobytes(), shape.geometry.id]
            if not iterator.next():
                break
        print("Done creating geometry")

    def create_id_map(self):
        if self.sql_type == "sqlite":
            statement = (
                "CREATE TABLE IF NOT EXISTS id_map (ifc_id integer PRIMARY KEY NOT NULL UNIQUE, ifc_class text);"
            )
        elif self.sql_type == "mysql":
            statement = """
            CREATE TABLE `id_map` (
              `ifc_id` int(10) unsigned NOT NULL,
              `ifc_class` varchar(255) NOT NULL,
              PRIMARY KEY (`ifc_id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb3 COLLATE=utf8mb3_general_ci;
            """
        self.c.execute(statement)

    def create_metadata(self):
        # There is no "standard" SQL serialisation, so we propose a convention
        # of a "metadata" table to hold high level metadata. This includes the
        # preprocessor field to uniquely identify the "variant" of SQL schema
        # used. If someone wants their own SQL schema variant, they can
        # identify it using the preprocessor field.
        # IfcOpenShell-1.0.0 represents a schema where 1 table = 1 declaration.
        # IfcOpenShell-2.0.0 represents a schema where tables represent types.
        metadata = ["IfcOpenShell-1.0.0", self.file.schema, self.file.header.file_description.description[0]]
        if self.sql_type == "sqlite":
            statement = "CREATE TABLE IF NOT EXISTS metadata (preprocessor text, schema text, mvd text);"
            self.c.execute(statement)
            self.c.execute("INSERT INTO metadata VALUES (?, ?, ?);", metadata)
        elif self.sql_type == "mysql":
            statement = """
            CREATE TABLE `metadata` (
              `preprocessor` varchar(255) NOT NULL,
              `schema` varchar(255) NOT NULL,
              `mvd` varchar(255) NOT NULL
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb3 COLLATE=utf8mb3_general_ci;
            """
            self.c.execute(statement)
            self.c.execute("INSERT INTO metadata VALUES (%s, %s, %s);", metadata)

    def create_pset_table(self):
        statement = """
        CREATE TABLE IF NOT EXISTS psets (
            ifc_id integer NOT NULL,
            pset_name text,
            name text,
            value text
        );
        """
        self.c.execute(statement)

    def create_geometry_table(self):
        statement = """
        CREATE TABLE IF NOT EXISTS shape (
            ifc_id integer NOT NULL,
            x real,
            y real,
            z real,
            matrix blob,
            geometry text
        );
        """
        self.c.execute(statement)

        statement = """
        CREATE TABLE IF NOT EXISTS geometry (
            id text NOT NULL,
            verts blob,
            edges blob,
            faces blob,
            material_ids blob,
            materials json
        );
        """

        if self.sql_type == "mysql":
            # mediumblob holds up to 16mb, longblob holds up to 4gb
            statement = statement.replace("blob", "mediumblob")

        self.c.execute(statement)

    def create_sqlite_table(self, ifc_class, declaration):
        statement = f"CREATE TABLE IF NOT EXISTS {ifc_class} ("

        if self.should_expand:
            statement += "ifc_id INTEGER NOT NULL"
        else:
            statement += "ifc_id INTEGER PRIMARY KEY NOT NULL UNIQUE"

        total_attributes = declaration.attribute_count()

        if total_attributes:
            statement += ","

        derived = declaration.derived()
        for i in range(0, total_attributes):
            attribute = declaration.attribute_by_index(i)
            primitive = ifcopenshell.util.attribute.get_primitive_type(attribute)
            if primitive in ("string", "enum"):
                data_type = "TEXT"
            elif primitive in ("entity", "integer", "boolean"):
                data_type = "INTEGER"
            elif primitive == "float":
                data_type = "REAL"
            elif self.should_expand and self.is_entity_list(attribute):
                data_type = "INTEGER"
            elif isinstance(primitive, tuple):
                data_type = "JSON"
            else:
                print(attribute, primitive)  # Not implemented?
            if not self.is_strict or derived[i]:
                optional = ""
            else:
                optional = "" if attribute.optional() else " NOT NULL"
            comma = "" if i == total_attributes - 1 else ","
            statement += f" `{attribute.name()}` {data_type}{optional}{comma}"
        if self.should_get_inverses:
            statement += ", inverses JSON"
        statement += ");"
        print(statement)
        self.c.execute(statement)

    def create_mysql_table(self, ifc_class, declaration):
        declaration = self.schema.declaration_by_name(ifc_class)
        statement = f"CREATE TABLE IF NOT EXISTS {ifc_class} ("
        statement += "`ifc_id` int(10) unsigned NOT NULL,"

        derived = declaration.derived()
        for attribute in declaration.all_attributes():
            primitive = ifcopenshell.util.attribute.get_primitive_type(attribute)
            if primitive in ("string", "enum"):
                if "IfcText" in str(attribute.type_of_attribute()):
                    data_type = "text"
                else:
                    data_type = "varchar(255)"
            elif primitive == "entity":
                data_type = "int(10) unsigned"
            elif primitive == "boolean":
                data_type = "tinyint(1)"
            elif primitive == "integer":
                data_type = "int(10)"
                if "Positive" in str(attribute.type_of_attribute()):
                    data_type += " unsigned"
            elif primitive == "float":
                data_type = "decimal(10,0)"
            elif self.should_expand and self.is_entity_list(attribute):
                data_type = "int(10) unsigned"
            elif isinstance(primitive, tuple):
                data_type = "JSON"
            else:
                print(attribute, primitive)  # Not implemented?
            if not self.is_strict or derived[i]:
                optional = "DEFAULT NULL"
            else:
                optional = "DEFAULT NULL" if attribute.optional() else "NOT NULL"
            statement += f" `{attribute.name()}` {data_type} {optional},"

        if self.should_expand:
            statement = statement[0:-1]
        else:
            statement += " PRIMARY KEY (`ifc_id`)"

        statement += ") ENGINE=InnoDB DEFAULT CHARSET=utf8mb3 COLLATE=utf8mb3_general_ci;"
        print(statement)
        self.c.execute(statement)

    def insert_data(self, ifc_class):
        print("Extracting data for", ifc_class)
        elements = self.file.by_type(ifc_class, include_subtypes=False)

        rows = []
        id_map_rows = []
        pset_rows = []

        for element in elements:
            nested_indices = []
            values = [element.id()]
            for i, attribute in enumerate(element):
                if isinstance(attribute, ifcopenshell.entity_instance):
                    if attribute.id():
                        values.append(attribute.id())
                    else:
                        values.append(json.dumps({"type": attribute.is_a(), "value": attribute.wrappedValue}))
                elif (
                    self.should_expand
                    and attribute
                    and isinstance(attribute, tuple)
                    and isinstance(attribute[0], ifcopenshell.entity_instance)
                ):
                    nested_indices.append(i + 1)
                    serialised_attribute = self.serialise_value(element, attribute)
                    if attribute[0].id():
                        values.append(serialised_attribute)
                    else:
                        values.append(json.dumps(serialised_attribute))
                elif isinstance(attribute, tuple):
                    attribute = self.serialise_value(element, attribute)
                    values.append(json.dumps(attribute))
                else:
                    values.append(attribute)

            if self.should_get_inverses:
                values.append(json.dumps([e.id() for e in self.file.get_inverse(element)]))

            if self.should_expand:
                rows.extend(self.get_permutations(values, nested_indices))
            else:
                rows.append(values)

            id_map_rows.append([element.id(), ifc_class])

            if self.should_get_psets:
                psets = ifcopenshell.util.element.get_psets(element)
                for pset_name, pset_data in psets.items():
                    for prop_name, value in pset_data.items():
                        if prop_name == "id":
                            continue
                        if isinstance(value, list):
                            value = json.dumps(value)
                        pset_rows.append([element.id(), pset_name, prop_name, value])

            if self.should_get_geometry:
                if element.id() not in self.shape_rows and getattr(element, "ObjectPlacement", None):
                    m = ifcopenshell.util.placement.get_local_placement(element.ObjectPlacement)
                    x, y, z = m[:, 3][0:3]
                    self.shape_rows[element.id()] = [element.id(), float(x), float(y), float(z), m.tobytes(), None]

        if self.sql_type == "sqlite":
            if rows:
                self.c.executemany(f"INSERT INTO {ifc_class} VALUES ({','.join(['?']*len(rows[0]))});", rows)
                self.c.executemany("INSERT INTO id_map VALUES (?, ?);", id_map_rows)
            if pset_rows:
                self.c.executemany("INSERT INTO psets VALUES (?, ?, ?, ?);", pset_rows)
        elif self.sql_type == "mysql":
            if rows:
                self.c.executemany(f"INSERT INTO {ifc_class} VALUES ({','.join(['%s']*len(rows[0]))});", rows)
                self.c.executemany("INSERT INTO id_map VALUES (%s, %s);", id_map_rows)
            if pset_rows:
                self.c.executemany("INSERT INTO psets VALUES (%s, %s, %s, %s);", pset_rows)

    def serialise_value(self, element, value):
        return element.walk(
            lambda v: isinstance(v, ifcopenshell.entity_instance),
            lambda v: v.id() if v.id() else {"type": v.is_a(), "value": v.wrappedValue},
            value,
        )

    def get_permutations(self, lst, indexes):
        nested_lists = [lst[i] for i in indexes]

        # Generate the Cartesian product of the nested lists
        products = list(itertools.product(*nested_lists))

        # Place the elements of each product back in their original positions
        final_lists = []
        for product in products:
            temp_list = lst[:]
            for i, index in enumerate(indexes):
                temp_list[index] = product[i]
            final_lists.append(temp_list)

        return final_lists

    def is_entity_list(self, attribute):
        attribute = str(attribute.type_of_attribute())
        if (attribute.startswith("<list") or attribute.startswith("<set")) and "<entity" in attribute:
            for data_type in re.findall("<(.*?) .*?>", attribute):
                if data_type not in ("list", "set", "select", "entity"):
                    return False
            return True
        return False



# IfcPatch - IFC patching utiliy
# Copyright (C) 2024 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.


import os
import re
import json
import time
import tempfile
import ifcopenshell
import ifcopenshell.util.element

try:
    import sqlite3
except:
    print("No SQLite support")


class Patcher:
    def __init__(
        self,
        src,
        file,
        logger,
    ):
        """Extracts properties and relationships from a IFC-SPF model to SQLite.

        This is a lossy extraction which simplifies popular properties to key
        value pairs.

        Example:

        .. code:: python

            result = ifcpatch.execute({"input": fn, "file": model, "recipe": "ExtractPropertiesToSQLite"})
            ifcpatch.write(result, "output.sqlite")
        """
        self.src = src
        self.file = file
        self.logger = logger

    def patch(self):
        tmp = tempfile.NamedTemporaryFile(delete=False)
        db_file = tmp.name
        self.db = sqlite3.connect(db_file)
        self.c = self.db.cursor()
        self.file_patched = db_file

        self.c.execute(
            """
            CREATE TABLE IF NOT EXISTS elements (
                id integer PRIMARY KEY NOT NULL UNIQUE,
                global_id text,
                ifc_class text,
                predefined_type text,
                name text,
                description text
            );
        """
        )
        self.c.execute("CREATE INDEX IF NOT EXISTS idx_global_id ON elements (global_id);")
        self.c.execute("CREATE INDEX IF NOT EXISTS idx_ifc_class ON elements (ifc_class);")
        self.c.execute("CREATE INDEX IF NOT EXISTS idx_predefined_type ON elements (predefined_type);")

        self.c.execute(
            """
            CREATE TABLE IF NOT EXISTS relationships (
                from_id integer NOT NULL,
                type text,
                to_id integer NOT NULL
            );
        """
        )
        self.c.execute("CREATE INDEX IF NOT EXISTS idx_from_id ON relationships (from_id);")

        self.c.execute(
            """
           CREATE TABLE IF NOT EXISTS properties (
               element_id integer NOT NULL,
               set_name text,
               name text,
               value text
           );
        """
        )
        self.c.execute("CREATE INDEX IF NOT EXISTS idx_element_id ON properties (element_id);")

        elements = self.file.by_type("IfcObjectDefinition")

        rows = []
        properties = []
        relationships = []
        id_map = {e.id(): i for i, e in enumerate(elements)}
        for i, element in enumerate(elements):
            rows.append(
                [
                    i,
                    element[0],  # IfcRoot.GlobalId
                    element.is_a(),
                    ifcopenshell.util.element.get_predefined_type(element),
                    element[2],  # IfcRoot.Name
                    element[3],  # IfcRoot.Description
                ]
            )
            psets = ifcopenshell.util.element.get_psets(element, should_inherit=False)
            for pset_name, pset_data in psets.items():
                for prop_name, value in pset_data.items():
                    if prop_name == "id" or value is None or value == "":
                        continue
                    if isinstance(value, bool):
                        value = "True" if value else "False"
                    elif not isinstance(value, str):
                        value = str(value)
                    properties.append([i, pset_name, prop_name, value])

            material = ifcopenshell.util.element.get_material(element, should_skip_usage=True)
            if material:
                name = getattr(material, "Name", getattr(material, "LayerSetName", None)) or "Unnamed"
                properties.append([i, "IFC Material", "Name", name])
                properties.append([i, "IFC Material", "Class", material.is_a()])
                if material.is_a("IfcMaterial"):
                    materials = []
                elif material.is_a("IfcMaterialLayerSet"):
                    for idx, item in enumerate(material.MaterialLayers):
                        material = item.Material
                        properties.append([i, "IFC Material", f"Layer {idx + 1} Name", getattr(item, "Name", None)])
                        properties.append([i, "IFC Material", f"Layer {idx + 1} Material", material.Name])
                        if category := getattr(material, "Category", None):
                            properties.append([i, "IFC Material", f"Layer {idx + 1} Category", category])
                elif material.is_a("IfcMaterialProfileSet"):
                    for idx, item in enumerate(material.MaterialProfiles):
                        material = item.Material
                        properties.append([i, "IFC Material", f"Profile {idx + 1} Name", item.Name])
                        properties.append([i, "IFC Material", f"Profile {idx + 1} Material", material.Name])
                        if category := getattr(material, "Category", None):
                            properties.append([i, "IFC Material", f"Profile {idx + 1} Category", category])
                elif material.is_a("IfcMaterialConstituentSet"):
                    for idx, item in enumerate(material.MaterialConstituents):
                        material = item.Material
                        properties.append([i, "IFC Material", f"Constituent {idx + 1} Name", item.Name])
                        properties.append([i, "IFC Material", f"Constituent {idx + 1} Material", material.Name])
                        if category := getattr(material, "Category", None):
                            properties.append([i, "IFC Material", f"Constituent {idx + 1} Category", category])
                elif material.is_a("IfcMaterialList"):
                    for idx, material in enumerate(material.Materials):
                        properties.append([i, "IFC Material", f"Material {idx + 1} Name", material.Name])
                        if category := getattr(material, "Category", None):
                            properties.append([i, "IFC Material", f"Material {idx + 1} Category", category])

            layers = ifcopenshell.util.element.get_layers(self.file, element)
            for idx, layer in enumerate(layers):
                properties.append([i, "IFC Presentation Layer Assignment", f"Layer {idx + 1}", layer.Name])

            relating_type = ifcopenshell.util.element.get_type(element)
            if relating_type and relating_type != element:
                relationships.append([i, "IfcRelDefinesByType", id_map[relating_type.id()]])

        self.c.executemany("INSERT INTO elements VALUES (?, ?, ?, ?, ?, ?);", rows)
        self.c.executemany("INSERT INTO properties VALUES (?, ?, ?, ?);", properties)
        self.c.executemany("INSERT INTO relationships VALUES (?, ?, ?);", relationships)

        self.db.commit()
        self.db.close()



# IfcPatch - IFC patching utiliy
# Copyright (C) 2020, 2021 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.

from collections import deque
import ifcopenshell.util.element


class Patcher:
    def __init__(self, src, file, logger):
        """Optimise the filesize of an IFC model by reusing non-rooted elements

        It is possible to non-losslessly optimise the filesize of an IFC model.

        Note that this is usually not recommended. Optimising runs a risk of
        losing some indirect semantic data critical for native IFC authoring.
        Most parties who recommend optimisation are not aware of these risks.
        Optimising is only safe in the context of read-only IFCs.

        If filesize is an issue, another approach would be to use IFCZIP
        instead to compress the model. Optimising the model only typically
        affects filesize and has minimal impact on load times. Large filesizes
        can usually be solved through other means. Consult the BlenderBIM Add-on
        documentation on dealing with large models for more details.

        This patch may be run multiple times with diminishing returns.

        Example:

        .. code:: python

            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "RecycleNonRootedElements", "arguments": []})
        """
        self.src = src
        self.file = file
        self.logger = logger

    def patch(self):
        deleted = []
        hashes = {}
        for element in self.file:
            if element.is_a("IfcRoot"):
                continue
            h = hash(tuple(element))
            if h in hashes:
                for inverse in self.file.get_inverse(element):
                    ifcopenshell.util.element.replace_attribute(inverse, element, hashes[h])
                deleted.append(element.id())
            else:
                hashes[h] = element
        deleted.sort()
        deleted_q = deque(deleted)
        new = ""
        for line in self.file.wrapped_data.to_string().split("\n"):
            try:
                if int(line.split("=")[0][1:]) != deleted_q[0]:
                    new += line + "\n"
                else:
                    deleted_q.popleft()
            except:
                new += line + "\n"
        self.file = new



# IfcPatch - IFC patching utiliy
# Copyright (C) 2023 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.

import ifcopenshell
import ifcopenshell.guid


class Patcher:
    def __init__(self, src, file, logger):
        """Convert nesting relationships to aggregate relationships

        Some software like Revit won't load nested children elements because
        they (incorrectly) don't consider it to be part of the spatial tree.
        For example, 12D software will use nesting.

        This patch converts all nest relationships into aggregate
        relationships.

        See bug: https://github.com/Autodesk/revit-ifc/issues/706

        Example:

        .. code:: python

            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "ConvertNestToAggregate", "arguments": []})
        """
        self.src = src
        self.file = file
        self.logger = logger

    def patch(self):
        for rel in self.file.by_type("IfcRelNests"):
            new = self.file.createIfcRelAggregates(
                ifcopenshell.guid.new(),
                rel.OwnerHistory,
                rel.Name,
                rel.Description,
                rel.RelatingObject,
                rel.RelatedObjects,
            )
            self.file.remove(rel)



# IfcPatch - IFC patching utiliy
# Copyright (C) 2020, 2021 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.

import ifcopenshell
import ifcopenshell.util.element
import ifcopenshell.util.unit
from typing import Union
from logging import Logger


class Patcher:
    def __init__(self, src: str, file: ifcopenshell.file, logger: Logger, filepath: str):
        """Merge two IFC models into one

        Note that other than combining the two IfcProject elements into one, no
        further processing will be done. This means that you may end up with
        duplicate spatial hierarchies (i.e. 2 sites, 2 buildings, etc).

        Will automatically convert length units in the second model to the main
        model's unit before merging.

        :param filepath: The filepath of the second IFC model to merge into the
            first. The first model is already specified as the input to
            IfcPatch.
        :type filepath: str

        Example:

        .. code:: python

            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "MergeProject", "arguments": ["/path/to/model2.ifc"]})
        """
        self.src = src
        self.file = file
        self.logger = logger
        self.filepath = filepath

    def patch(self):
        source = ifcopenshell.open(self.filepath)
        # make sure models units will match
        if (main_unit := self.get_unit_name(self.file)) != self.get_unit_name(source):
            source = ifcopenshell.util.unit.convert_file_length_units(source, main_unit)

        self.existing_contexts: list[ifcopenshell.entity_instance] = self.file.by_type(
            "IfcGeometricRepresentationContext"
        )
        self.added_contexts: set[ifcopenshell.entity_instance] = set()

        original_project = self.file.by_type("IfcProject")[0]
        merged_project = self.file.add(source.by_type("IfcProject")[0])

        for element in source.by_type("IfcGeometricRepresentationContext"):
            new = self.file.add(element)
            self.added_contexts.add(new)

        for element in source:
            self.file.add(element)

        for inverse in self.file.get_inverse(merged_project):
            ifcopenshell.util.element.replace_attribute(inverse, merged_project, original_project)
        self.file.remove(merged_project)

        self.reuse_existing_contexts()

    def get_unit_name(self, ifc_file: ifcopenshell.file) -> str:
        length_unit = ifcopenshell.util.unit.get_project_unit(ifc_file, "LENGTHUNIT")
        return ifcopenshell.util.unit.get_full_unit_name(length_unit)

    def reuse_existing_contexts(self):
        to_delete = set()

        for added_context in self.added_contexts:
            equivalent_existing_context = self.get_equivalent_existing_context(added_context)
            if equivalent_existing_context:
                for inverse in self.file.get_inverse(added_context):
                    ifcopenshell.util.element.replace_attribute(inverse, added_context, equivalent_existing_context)
                to_delete.add(added_context)

        for added_context in to_delete:
            ifcopenshell.util.element.remove_deep2(self.file, added_context)

    def get_equivalent_existing_context(
        self, added_context: ifcopenshell.entity_instance
    ) -> Union[ifcopenshell.entity_instance, None]:
        for context in self.existing_contexts:
            if context.is_a() != added_context.is_a():
                continue
            if context.is_a("IfcGeometricRepresentationSubContext"):
                if (
                    context.ContextType == added_context.ContextType
                    and context.ContextIdentifier == added_context.ContextIdentifier
                    and context.TargetView == added_context.TargetView
                ):
                    return context
            elif (
                context.ContextType == added_context.ContextType
                and context.ContextIdentifier == added_context.ContextIdentifier
            ):
                return context



# IfcPatch - IFC patching utiliy
# Copyright (C) 2023 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.


import os
import math
import ifcopenshell
import ifcopenshell.geom
import ifcopenshell.util.unit
import ifcopenshell.util.schema
import ifcopenshell.util.element


class Patcher:
    def __init__(self, src, file, logger):
        """Fix missing door swings in Revit when viewing ArchiCAD IFCs

        ArchiCAD has the ability to store 2D data with objects like doors for
        door swings. ArchiCAD's implementation is not 100% correct (using
        footprint instead of annotation contexts), but otherwise not too
        shabby.

        Revit, however, is incapable of understanding this 2D representation.
        Revit users linking in IFCs produced by ArchiCAD may experience the
        following symptoms:

        A. Invisible doors, and difficulty selecting doors
        B. Invisible door swings, or only visible at particular view ranges
        C. Weird arc shapes around doors
        D. Extra lines around doors and walls
        E. Cannot easily change visibility graphics of 2D vs 3D elements
        F. Cannot view 2D data in a 3D view

        This is caused by the perfect storm of Revit IFC bugs, which we will
        work through methodically. For programmers interested in the details of
        how we fix this, read the comments of the patch function.

        Example:

        .. code:: python

            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "FixArchiCADToRevitDoorSwings", "arguments": []})
        """
        self.src = src
        self.file = file
        self.logger = logger

    def patch(self):
        # Revit has the ability to switch between 3D representations and 2D
        # representations (e.g. in plan view). It does this by detecting IFC
        # representations that belong to either the Model Body representation
        # for 3D and the "Annotation" representation for 2D. Note: this is not a real
        # "Annotation" representation. It is an invalid representation that is
        # hardcoded in Revit. Nevertheless, if you follow their incorrect
        # context naming convention by renaming ArchiCAD's also incorrect
        # FootPrint context (but less incorrect) to Annotation, Revit will show
        # 2D door representations as desired and switch between 2D and 3D
        # intelligently.
        #
        # See bug https://github.com/Autodesk/revit-ifc/issues/187

        # for context in self.file.by_type("IfcGeometricRepresentationSubContext"):
        #    if context.ContextIdentifier == "FootPrint":
        #        context.ContextIdentifier = None
        #        context.ContextType = "Annotation"
        #        context.ParentContext = [
        #            c
        #            for c in self.file.by_type("IfcGeometricRepresentationContext", include_subtypes=False)
        #            if c.ContextType == "Plan"
        #        ][0]

        # However, we will not use this approach (hence it is commented out).
        # This is because Revit will generate a single object with both a
        # magically switching 2D and 3D representation - you cannot control both
        # representations independently. Furthermore, the 2D representation is
        # always generated at a Z value of 0. This means that the door swings
        # are always invisible on any view range apart from those typically near
        # an elevation of 0 which won't work for most projects. Because they are
        # merged into a single object, you cannot fix the 2D elevation without
        # breaking your 3D elevation. Even if we split it into 2D and 3D
        # objects, the 2D object can only be moved up in 3D or section, but
        # Revit makes the object invisible in 3D or section, so go fish.
        #
        # See symptom B.

        # Purge axis representations from walls. Revit will draw the axis (a
        # line through the wall) but doesn't give you control over its
        # visibility, resulting in very ugly walls and lines typically cutting
        # across door openings.
        #
        # See bug https://github.com/Autodesk/revit-ifc/issues/360
        #
        # This fixes symptom D.

        for wall in self.file.by_type("IfcWall"):
            reps = list(wall.Representation.Representations)
            reps_excluding_axis = [r for r in reps if r.RepresentationIdentifier != "Axis"]
            wall.Representation.Representations = reps_excluding_axis

        # We're going to split the 3D door body and the 2D door swing into two
        # objects. Before we do that, we're going to cull the 2D representation
        # from door types. If we didn't, this scenario may occur:
        #
        # 1. 3D IfcDoors with IfcDoorType import as Revit doors
        # 2. 2D IfcBuildingElementProxy with no type import as Revit generic model
        # 3. The Revit user turns off the visibility of the door category.
        # 4. The user is shocked to see that generic models are now invisible,
        #    even though it is clearly a separate category to doors.
        #
        # This should be a warning to all Revit users that unintuitively
        # visibility graphics actually do _not_ affect the categories (e.g.
        # Door, Window) for instances. Instead, they target categories for
        # representation maps types, even if an instance is untyped. Yikes!
        #
        # See bug: https://github.com/Autodesk/revit-ifc/issues/362
        #
        # This fixes symptom E.

        for door in self.file.by_type("IfcDoorType"):
            rep_maps = list(door.RepresentationMaps or [])
            door.RepresentationMaps = [
                r
                for r in rep_maps
                if r.MappedRepresentation
                if r.MappedRepresentation.RepresentationIdentifier != "FootPrint"
            ]

        # Let's now duplicate each door. The original door shall only retain its
        # non 2D representations. The copied door will be changed into a
        # "IfcDiscreteAccessory" so it can be filtered separately in Revit.
        # Then, the copied door will only retain 2D representations. However,
        # we're going to change the 2D representation context into a Model Body
        # 3D context. This tricks Revit into converting the 2D data into 3D
        # model line geometry (or something similar to that).
        #
        # See bug https://github.com/Autodesk/revit-ifc/issues/358
        #
        # This fixes symptom A, B, and F.

        for door in self.file.by_type("IfcDoor"):
            reps = list(door.Representation.Representations)
            reps_excluding_footprint = [r for r in reps if r.RepresentationIdentifier != "FootPrint"]
            footprint_reps = [r for r in reps if r.RepresentationIdentifier == "FootPrint"]
            if not footprint_reps:
                continue
            door_copy = ifcopenshell.util.element.copy(self.file, door)
            door_copy = ifcopenshell.util.schema.reassign_class(self.file, door_copy, "IfcDiscreteAccessory")
            door_copy.Representation = ifcopenshell.util.element.copy(self.file, door.Representation)
            door_copy.Representation.Representations = footprint_reps
            door.Representation.Representations = reps_excluding_footprint
            related_elements = list(door.ContainedInStructure[0].RelatedElements)
            related_elements.append(door_copy)
            door.ContainedInStructure[0].RelatedElements = related_elements

            body_context = [
                c for c in self.file.by_type("IfcGeometricRepresentationSubContext") if c.ContextIdentifier == "Body"
            ][0]
            for subelement in self.file.traverse(footprint_reps[0]):
                if not subelement.is_a("IfcShapeRepresentation"):
                    continue
                subelement.ContextOfItems = body_context
                subelement.RepresentationIdentifier = "Body"
                subelement.RepresentationType = "Curve3D"

        # Revit has bugs with handling arc index, which is relevant in IFC4. Our
        # solution is to facet the arcs and swap out the polycurves with new
        # faceted polycurves.
        #
        # See bug https://github.com/Autodesk/revit-ifc/issues/359
        #
        # This fixes symptom C.

        settings = ifcopenshell.geom.settings()
        settings.set(settings.INCLUDE_CURVES, True)
        unit_scale = ifcopenshell.util.unit.calculate_unit_scale(self.file)
        for curve in self.file.by_type("IfcIndexedPolyCurve"):
            if True in [s.is_a("IfcArcIndex") for s in curve.Segments]:
                shape = ifcopenshell.geom.create_shape(settings, curve)
                e = shape.edges
                v = shape.verts
                is_2d = curve.Points.is_a("IfcCartesianPointList2D")
                if is_2d:
                    vertices = [[v[i] / unit_scale, v[i + 1] / unit_scale] for i in range(0, len(v), 3)]
                else:
                    vertices = [
                        [v[i] / unit_scale, v[i + 1] / unit_scale, v[i + 2] / unit_scale] for i in range(0, len(v), 3)
                    ]
                edges = [[e[i] + 1, e[i + 1] + 1] for i in range(0, len(e), 2)]
                points = []
                curve.Points = f.create_entity(curve.Points.is_a(), vertices)
                curve.Segments = [f.createIfcLineIndex(e) for e in edges]



# IfcPatch - IFC patching utiliy
# Copyright (C) 2023 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import ifcopenshell
from logging import Logger


class Patcher:
    def __init__(self, src: str, file: ifcopenshell.file, logger: Logger):
        """Purge IFC properties, relationships, and other data

        In some rare cases (i.e. "resetting" a model or for security purposes)
        you may need to purge all relationships and data within a model but
        retain the ability to view the geometry in a model. This patch purges
        the following relationships:

        - STEP comments, which may include software export metadata
        - Header data, which may include filenames
        - Owner, person, organisation, and application data
        - Addresses
        - Rooted names
        - Tags (that reference original software)
        - Properties and quantities
        - Materials
        - Styles (e.g. colours)
        - External references like classifications, documents, libraries
        - Types
        - Groups (such as model groups or search groups) and systems
        - Profile names

        Example:

        .. code:: python

            # Watch the world burn
            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "PurgeData", "arguments": []})
        """
        self.src = src
        self.file = file
        self.logger = logger

    def patch(self):
        self.file.wrapped_data.header.file_name.name = "Rabbit"
        self.file.wrapped_data.header.file_name.time_stamp = (
            datetime.datetime.utcnow()
            .replace(tzinfo=datetime.timezone.utc)
            .astimezone()
            .replace(microsecond=0)
            .isoformat()
        )
        self.file.wrapped_data.header.file_name.preprocessor_version = "Rabbit"
        self.file.wrapped_data.header.file_name.originating_system = "Rabbit"

        for element in self.file.by_type("IfcApplication"):
            element.Version = "Rabbit"
            element.ApplicationFullName = "Rabbit"
            element.ApplicationIdentifier = "Rabbit"

        for element in self.file.by_type("IfcOrganization"):
            element[0] = None  # Id
            element.Name = "Rabbit"
            element.Description = None
            element.Roles = []
            element.Addresses = []

        for element in self.file.by_type("IfcPerson"):
            for i in range(0, len(element)):
                element[i] = None

        for element in self.file.by_type("IfcPersonAndOrganization"):
            element.Roles = []

        for element in self.file.by_type("IfcActorRole"):
            self.file.remove(element)

        for element in self.file.by_type("IfcAddress"):
            self.file.remove(element)

        for element in self.file.by_type("IfcRoot"):
            element.Name = "Rabbit"
            element.Description = "Rabbit"

        if self.file.schema == "IFC2X3":
            elements = self.file.by_type("IfcProject")
        else:
            elements = self.file.by_type("IfcContext")
        for element in elements:
            element.ObjectType = None
            element.LongName = None
            element.Phase = None

        if self.file.schema == "IFC2X3":
            elements = self.file.by_type("IfcSpatialStructureElement")
        else:
            elements = self.file.by_type("IfcSpatialElement")
        for element in elements:
            element.LongName = None

        for element in self.file.by_type("IfcOwnerHistory"):
            element.State = None
            element.ChangeAction = "NOCHANGE" if self.file.schema == "IFC2X3" else None
            element.LastModifiedDate = None
            element.LastModifyingUser = None
            element.LastModifyingApplication = None
            element.CreationDate = 0

        for element in self.file.by_type("IfcRelDefines"):
            self.file.remove(element)

        for element in self.file.by_type("IfcTypeProduct"):
            self.file.remove(element)

        for element in self.file.by_type("IfcElement"):
            element.Tag = None
            element.ObjectType = None

        for element in self.file.by_type("IfcPropertyDefinition"):
            self.file.remove(element)

        if self.file.schema == "IFC2X3":
            elements = self.file.by_type("IfcProperty")
        else:
            elements = self.file.by_type("IfcPropertyAbstraction")
        for element in elements:
            self.file.remove(element)

        for element in self.file.by_type("IfcPhysicalQuantity"):
            self.file.remove(element)

        for element in self.file.by_type("IfcRelAssociates"):
            self.file.remove(element)

        if self.file.schema == "IFC2X3":
            elements = (
                self.file.by_type("IfcMaterial")
                + self.file.by_type("IfcMaterialLayer")
                + self.file.by_type("IfcMaterialLayerSet")
                + self.file.by_type("IfcMaterialLayerSetUsage")
                + self.file.by_type("IfcMaterialList")
            )
        else:
            elements = self.file.by_type("IfcMaterialDefinition")
        for element in elements:
            self.file.remove(element)

        for element in self.file.by_type("IfcMaterialDefinitionRepresentation"):
            self.file.remove(element)

        if self.file.schema == "IFC2X3":
            elements = []
        else:
            elements = self.file.by_type("IfcMaterialUsageDefinition")
        for element in elements:
            self.file.remove(element)

        for element in self.file.by_type("IfcMaterialList"):
            self.file.remove(element)

        for element in self.file.by_type("IfcStyledItem"):
            self.file.remove(element)

        for element in self.file.by_type("IfcStyledRepresentation"):
            self.file.remove(element)

        for element in self.file.by_type("IfcPresentationStyle"):
            self.file.remove(element)

        if self.file.schema == "IFC2X3":
            elements = (
                self.file.by_type("IfcColourSpecification")
                + self.file.by_type("IfcCurveStyleFont")
                + self.file.by_type("IfcCurveStyleFontAndScaling")
                + self.file.by_type("IfcCurveStyleFontPattern")
                + self.file.by_type("IfcPreDefinedItem")
                + self.file.by_type("IfcSurfaceStyleLighting")
                + self.file.by_type("IfcSurfaceStyleRefraction")
                + self.file.by_type("IfcSurfaceStyleShading")
                + self.file.by_type("IfcSurfaceStyleWithTextures")
                + self.file.by_type("IfcSurfaceTexture")
                + self.file.by_type("IfcTextStyleForDefinedFont")
                + self.file.by_type("IfcTextStyleTextModel")
                + self.file.by_type("IfcTextStyleWithBoxCharacteristics")
                + self.file.by_type("IfcTextureCoordinate")
                + self.file.by_type("IfcTextureVertex")
                + self.file.by_type("IfcTextStyleForDefinedFont")
                + self.file.by_type("IfcTextStyleForDefinedFont")
            )
        else:
            elements = self.file.by_type("IfcPresentationItem")
        for element in elements:
            self.file.remove(element)

        if self.file.schema == "IFC2X3":
            elements = (
                self.file.by_type("IfcClassification")
                + self.file.by_type("IfcDocumentInformation")
                + self.file.by_type("IfcLibraryInformation")
            )
        else:
            elements = self.file.by_type("IfcExternalInformation")
        for element in elements:
            self.file.remove(element)

        for element in self.file.by_type("IfcExternalReference"):
            self.file.remove(element)

        for element in self.file.by_type("IfcTypeObject"):
            self.file.remove(element)

        for element in self.file.by_type("IfcGroup"):
            self.file.remove(element)

        for element in self.file.by_type("IfcRelAssigns"):
            self.file.remove(element)

        for element in self.file.by_type("IfcProfileDef"):
            element.ProfileName = None

        for element in self.file.by_type("IfcShapeAspect"):
            element.Name = None
            element.Description = None

        for element in self.file.by_type("IfcPresentationLayerAssignment"):
            self.file.remove(element)



# IfcPatch - IFC patching utiliy
# Copyright (C) 2020, 2021 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.


class Patcher:
    def __init__(self, src, file, logger):
        """Removes any 3D geometry associated with a site or multiple sites

        If no sites or no site geometry is present, nothing happens.

        Example:

        .. code:: python

            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "RemoveSiteRepresentation", "arguments": []})
        """
        self.src = src
        self.file = file
        self.logger = logger

    def patch(self):
        project = self.file.by_type("IfcProject")[0]
        sites = self.find_decomposed_ifc_class(project, "IfcSite")
        for site in sites:
            site.Representation = None

    def find_decomposed_ifc_class(self, element, ifc_class):
        results = []
        rel_aggregates = element.IsDecomposedBy
        if not rel_aggregates:
            return results
        for rel_aggregate in rel_aggregates:
            for part in rel_aggregate.RelatedObjects:
                if part.is_a(ifc_class):
                    results.append(part)
                results.extend(self.find_decomposed_ifc_class(part, ifc_class))
        return results



# IfcOpenShell - IFC toolkit and geometry engine
# Copyright (C) 2022 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcOpenShell.
#
# IfcOpenShell is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcOpenShell is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcOpenShell.  If not, see <http://www.gnu.org/licenses/>.

import os
import pytest
import ifcpatch
import ifcopenshell
import ifcopenshell.api
import ifcopenshell.util.element
import test.bootstrap


class TestExtractElements(test.bootstrap.IFC4):
    def test_basic(self):
        project = ifcopenshell.api.run("root.create_entity", self.file, ifc_class="IfcProject")
        wall = ifcopenshell.api.run("root.create_entity", self.file, ifc_class="IfcWall")
        output = ifcpatch.execute({"file": self.file, "recipe": "ExtractElements", "arguments": ["IfcWall"]})

        assert output.by_type("IfcProject")[0].GlobalId == project.GlobalId
        assert output.by_type("IfcWall")[0].GlobalId == wall.GlobalId

    def test_keep_spatial_structure(self):
        project = ifcopenshell.api.run("root.create_entity", self.file, ifc_class="IfcProject")

        site = ifcopenshell.api.run("root.create_entity", self.file, ifc_class="IfcSite")
        building = ifcopenshell.api.run("root.create_entity", self.file, ifc_class="IfcBuilding")
        storey = ifcopenshell.api.run("root.create_entity", self.file, ifc_class="IfcBuildingStorey")
        wall = ifcopenshell.api.run("root.create_entity", self.file, ifc_class="IfcWall")
        ifcopenshell.api.run("aggregate.assign_object", self.file, products=[building], relating_object=site)
        ifcopenshell.api.run("aggregate.assign_object", self.file, products=[storey], relating_object=building)
        ifcopenshell.api.run("spatial.assign_container", self.file, products=[wall], relating_structure=storey)

        output = ifcpatch.execute({"file": self.file, "recipe": "ExtractElements", "arguments": ["IfcWall"]})

        wall_new = output.by_type("IfcWall")[0]
        assert (storey_new := ifcopenshell.util.element.get_container(wall_new)).GlobalId == storey.GlobalId
        assert (building_new := ifcopenshell.util.element.get_aggregate(storey_new)).GlobalId == building.GlobalId
        assert (site_new := ifcopenshell.util.element.get_aggregate(building_new)).GlobalId == site.GlobalId

    def test_keep_aggregate_in_spatial_structure(self):
        project = ifcopenshell.api.run("root.create_entity", self.file, ifc_class="IfcProject")

        element = ifcopenshell.api.run("root.create_entity", self.file, ifc_class="IfcElementAssembly")
        container = ifcopenshell.api.run("root.create_entity", self.file, ifc_class="IfcBuildingStorey")
        subelement = ifcopenshell.api.run("root.create_entity", self.file, ifc_class="IfcWall")
        ifcopenshell.api.run("spatial.assign_container", self.file, products=[element], relating_structure=container)
        ifcopenshell.api.run("aggregate.assign_object", self.file, products=[subelement], relating_object=element)

        output = ifcpatch.execute({"file": self.file, "recipe": "ExtractElements", "arguments": ["IfcWall"]})

        wall_new = output.by_type("IfcWall")[0]
        assembly = output.by_type("IfcElementAssembly")[0]

        assert ifcopenshell.util.element.get_aggregate(wall_new).GlobalId == element.GlobalId
        assert ifcopenshell.util.element.get_container(assembly).GlobalId == container.GlobalId

    def test_getting_the_psets_of_a_product_as_a_dictionary(self):
        ifc = ifcopenshell.open(os.path.join(os.getcwd(), "test", "files", "basic.ifc"))
        output = ifcpatch.execute({"file": ifc, "recipe": "ExtractElements", "arguments": ["IfcWall"]})
        assert output.by_type("IfcWall")
        assert not output.by_type("IfcSlab")


class TestExtractElementsIFC2X3(test.bootstrap.IFC2X3, TestExtractElements):
    pass



# IfcPatch - IFC patching utiliy
# Copyright (C) 2023 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.


class Patcher:
    def __init__(self, src, file, logger):
        """Removes the built-in Revit Uniformat classification.

        Revit has a bug (see https://github.com/Autodesk/revit-ifc/issues/486)
        where it always inserts a Uniformat classification regardless if your
        project needs it or not. This patch removes it.

        Example:

        .. code:: python

            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "RemoveRevitUniformatClassification"})
        """
        self.src = src
        self.file = file
        self.logger = logger

    def patch(self):
        for classification in self.file.by_type("IfcClassification"):
            if classification.Name != "Uniformat":
                continue
            references = self.get_references(classification)
            for reference in references:
                self.file.remove(reference)
            self.file.remove(classification)
            for rel in self.file.by_type("IfcRelAssociatesClassification"):
                if not rel.RelatingClassification:
                    self.file.remove(rel)
            for rel in self.file.by_type("IfcExternalReferenceRelationship"):
                if not rel.RelatingReference:
                    self.file.remove(rel)

    def get_references(self, classification):
        results = []
        if not classification.HasReferences:
            return results
        for reference in classification.HasReferences:
            results.append(reference)
            results.extend(self.get_references(reference))
        return results



#!/usr/bin/env python3

# IfcPatch - IFC patching utiliy
# Copyright (C) 2020, 2021, 2022 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import ifcpatch
import ifcopenshell

parser = argparse.ArgumentParser(description="Patches IFC files to fix badly formatted data")
parser.add_argument("-i", "--input", type=str, required=True, help="The IFC file to patch")
parser.add_argument("-o", "--output", type=str, help="The output file to save the patched IFC")
parser.add_argument("-r", "--recipe", type=str, required=True, help="Name of the recipe to use when patching")
parser.add_argument("-l", "--log", type=str, help="Specify a log file", default="ifcpatch.log")
parser.add_argument("-a", "--arguments", nargs="+", help="Specify custom arguments to the patch recipe")
args = vars(parser.parse_args())

print("# Loading IFC file ...")
args["file"] = ifcopenshell.open(args["input"])

print("# Patching ...")
output = ifcpatch.execute(args)

print("# Writing patched file ...")
if not args["output"]:
    args["output"] = args["input"]
ifcpatch.write(output, args["output"])

print("# All tasks are complete :-)")
